"""
.. module:: teamcontrol.scm
   :platform: Unix, Windows
   :synopsis: Source Code Manager related tools

.. moduleauthor:: Caio Romao <caioromao@gmail.com>

"""

__all__ = ["Git", "SCM"]

import re
import subprocess

from abc import ABCMeta, abstractmethod, abstractproperty
from teamcontrol import ENCODING


class SCMConfigurationError(Exception):
    pass


class SCM(object):

    __metaclass__ = ABCMeta

    def run_command(self, cmd):
        pipe = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        stdout, _ = pipe.communicate()
        # Convert read data to Unicode:
        # * In py3k this goes from byte to str
        # * In py2k it goes from str to unicode
        return stdout.decode(ENCODING).strip()

    @classmethod
    def create_manager(cls):
        # TODO: Implement SCM discovery heuristics?
        return Git()

    @abstractproperty
    def parser(self):
        pass

    @abstractmethod
    def list_changed_files(self, data):
        pass

    @abstractmethod
    def get_user_settings(self):
        pass

    @abstractmethod
    def should_ignore(self, data):
        pass

    def get_files(self, input_data):
        files = set()
        for data in input_data:
            if self.should_ignore(data):
                continue

            parsed_data = self.parser.parse(data)
            files |= self.list_changed_files(parsed_data)
        return files


class Git(SCM):

    LS_FILES_CMD = r'git show --pretty=format: --name-only {0}'
    GET_TEAM_CMD = r'git config teamcontrol.team'
    GET_CONFIG_CMD = r'git config teamcontrol.config'

    IGNORE_RE = re.compile(r'^0+-')

    NO_TEAM_ERR = ("No configuration found for teamcontrol.team.\n"
                   "Set your team name by issuing:\n"
                   "$ git config teamcontrol.team MY_TEAM_NAME")
    NO_CONFIG_ERR = ("No configuration found for teamcontrol.config.\n"
                     "Set your configuration path by issuing:\n"
                     "$ git config teamcontrol.config /path/to/config")

    def __init__(self):
        from teamcontrol.parser import MboxParser
        self.__parser = MboxParser()

    @property
    def parser(self):
        return self.__parser

    def should_ignore(self, filename):
        if self.IGNORE_RE.match(filename):
            return True
        return False

    def list_changed_files(self, sha):
        cmd = self.LS_FILES_CMD.format(sha)
        stdout = self.run_command(cmd)

        return set(line.strip() for line in stdout.split())

    def get_user_settings(self):
        teamname = self.run_command(self.GET_TEAM_CMD)
        configpath = self.run_command(self.GET_CONFIG_CMD)

        if not teamname:
            raise SCMConfigurationError(self.NO_TEAM_ERR)

        if not configpath:
            raise SCMConfigurationError(self.NO_CONFIG_ERR)

        return dict(team=teamname, config=configpath)
